/**
       This file is part of Adguard for iOS (https://github.com/AdguardTeam/AdguardForiOS).
       Copyright © Adguard Software Limited. All rights reserved.
 
       Adguard for iOS is free software: you can redistribute it and/or modify
       it under the terms of the GNU General Public License as published by
       the Free Software Foundation, either version 3 of the License, or
       (at your option) any later version.
 
       Adguard for iOS is distributed in the hope that it will be useful,
       but WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
       GNU General Public License for more details.
 
       You should have received a copy of the GNU General Public License
       along with Adguard for iOS.  If not, see <http://www.gnu.org/licenses/>.
 */

import Foundation

// MARK - service protocol -

/**
 DnsProvidersService is responsible for managing dns providers
 */
@objc
protocol DnsProvidersServiceProtocol {
    var vpnManager: VpnManagerProtocol? { get set }
    
    var allProviders: [DnsProviderInfo] { get }
    var predefinedProviders: [DnsProviderInfo] { get }
    var customProviders: [DnsProviderInfo] { get }
    var adguardDohServer: DnsServerInfo? { get }
    var adguardFamilyDohServer: DnsServerInfo? { get }
    
    var activeDnsServer: DnsServerInfo? { get set }
    var activeDnsProvider: DnsProviderInfo? { get }
    var currentServerName: String { get }
    
    func addCustomProvider(name: String, upstream: String)->DnsProviderInfo
    func deleteProvider(_ provider: DnsProviderInfo)
    func updateProvider(_ provider: DnsProviderInfo)
    func isCustomProvider(_ provider: DnsProviderInfo)->Bool
    func isCustomServer(_ server: DnsServerInfo)->Bool
    func isActiveProvider(_ provider: DnsProviderInfo)->Bool
    func getServer(serverId: Int)->DnsServerInfo?
    func getServerName(serverId: Int)->String?
    
    func reset()
}

@objc class DnsProvidersService: NSObject, DnsProvidersServiceProtocol {
    
    private var predefinedProvidersInternal: [DnsProviderInfo]?
    private var customProvidersInternal: [DnsProviderInfo]?
    private let workingQueue = DispatchQueue(label: "dns providers queue")
    
    @objc private let resources: AESharedResourcesProtocol
    weak var vpnManager: VpnManagerProtocol?
    
    private let APDefaultsCustomDnsProviders = "APDefaultsCustomDnsProviders"
    
    
    /*
     Language map for providers_i18n.json
     Some languages codes from json differ from apple ones
    */
    private var languageMap = ["es": "es-ES",
                               "zh-Hans": "zh-CN",
                               "zh-Hant:": "zh-TW"]
    
    private var currentLocaleCode: String
    
    private let systemDefaultProviderId = 10000
    
    @objc convenience init(resources: AESharedResourcesProtocol) {
        self.init(resources: resources, locale: Bundle.main.preferredLocaleCode)
    }
    
    @objc init(resources: AESharedResourcesProtocol, locale: String) {
        self.resources = resources
        
        // migration:
        // in app version 3.1.4 and below we mistakenly used the name Adguard.DnsProviderInfo with namespace
        // now we use DnsProviderInfo
        NSKeyedUnarchiver.setClass(DnsProviderInfo.self, forClassName: "Adguard.DnsProviderInfo")
        
        self.currentLocaleCode = locale
        
        super.init()
        
        self.migrateActiveServerIfNeeded()
    }
    
    @objc var predefinedProviders: [DnsProviderInfo] {
        get {
            if predefinedProvidersInternal == nil {
                self.initializeDnsProviders()
            }
            
            return predefinedProvidersInternal ?? []
        }
    }
    
    var customProviders: [DnsProviderInfo] {
        get {
            if customProvidersInternal == nil {
                if let data = resources.sharedDefaults().object(forKey: APDefaultsCustomDnsProviders) as? Data {
                    customProvidersInternal = NSKeyedUnarchiver.unarchiveObject(with: data) as? [DnsProviderInfo] ?? []
                }
            }
            
            return customProvidersInternal ?? []
        }
        
        set {
            willChangeValue(for: \.allProviders)
            customProvidersInternal = newValue
            if customProvidersInternal != nil {
                let data = NSKeyedArchiver.archivedData(withRootObject: customProvidersInternal!)
                resources.sharedDefaults().set(data, forKey: APDefaultsCustomDnsProviders)
                resources.sharedDefaults().synchronize()
            }
            didChangeValue(for: \.allProviders)
        }
    }
    
    @objc var allProviders: [DnsProviderInfo] {
        return predefinedProviders + customProviders
    }
    
    @objc var activeDnsServer: DnsServerInfo? {
        get {
            guard let data = resources.sharedDefaults().object(forKey:AEDefaultsActiveDnsServer) as? Data else { return nil }
            return NSKeyedUnarchiver.unarchiveObject(with: data) as? DnsServerInfo
            
        }
        set {
            guard let server = newValue else {
                resources.sharedDefaults().removeObject(forKey: AEDefaultsActiveDnsServer)
                return
            }
            
            let data = NSKeyedArchiver.archivedData(withRootObject:server)
            
            willChangeValue(for: \.activeDnsServer)
            resources.sharedDefaults().set(data, forKey: AEDefaultsActiveDnsServer)
            didChangeValue(for: \.activeDnsServer)
        }
    }
    
    var activeDnsProvider: DnsProviderInfo? {
        get {
            let activeServer = activeDnsServer
            return allProviders.first { (provider) in
                provider.servers?.contains { $0.serverId == activeServer?.serverId } ?? false
            }
        }
    }
    
    /* True if there is no saved server id among fetched providers */
    private var providerIsMissing: Bool { activeDnsProvider == nil && activeDnsServer != nil }
    
    func addCustomProvider(name: String, upstream: String) -> DnsProviderInfo {
        let provider = DnsProviderInfo(name: name)
        
        let server = DnsServerInfo(dnsProtocol: .dns, serverId: UUID().uuidString, name: name, upstreams: [upstream])
        
        server.dnsProtocol = DnsProtocol.getProtocolByUpstream(upstream)
        provider.servers = [server]
        
        workingQueue.sync { [weak self] in
            self?.customProviders.append(provider)
        }
        
        return provider
    }
    
    func deleteProvider(_ provider: DnsProviderInfo)  {
        
        workingQueue.sync { [weak self] in
            guard let self = self else { return }

            self.willChangeValue(forKey: "allProviders")
            
            // search provider by server id.
            customProviders = customProviders.filter {
                $0.servers?.first?.serverId != provider.servers?.first?.serverId
            }
            
            self.didChangeValue(forKey: "allProviders")
        }
    }
    
    func updateProvider(_ provider: DnsProviderInfo)  {
        workingQueue.sync { [weak self] in
            guard let self = self else { return }

            self.willChangeValue(forKey: "allProviders")
            
            // search provider by server id.
            customProviders = customProviders.map { (currentProvider)->DnsProviderInfo in
                guard let server = currentProvider.servers?.first else { return currentProvider }
                if server.serverId == provider.servers?.first?.serverId {
                    server.name = provider.servers?.first?.name ?? ""
                    server.upstreams = provider.servers?.first?.upstreams ?? []
                    server.dnsProtocol = provider.servers?.first?.dnsProtocol ?? .dns
                    
                    if server.serverId == activeDnsServer?.serverId {
                        activeDnsServer = server
                    }
                }
                
                return currentProvider
            }
            
            self.didChangeValue(forKey: "allProviders")
        }
    }
    
    func isCustomProvider(_ provider: DnsProviderInfo)->Bool {
        return customProviders.contains {
            $0.servers?.first?.serverId == provider.servers?.first?.serverId
        }
    }
    
    @objc
    func isCustomServer(_ server: DnsServerInfo)->Bool {
        return customProviders.contains {
            $0.servers?.first?.serverId == server.serverId
        }
    }
    
    func isActiveProvider(_ provider: DnsProviderInfo)->Bool {
        guard let server = activeDnsServer else { return false }
        return provider.servers?.contains { $0.serverId == server.serverId } ?? false
    }
    
    var currentServerName: String {
        guard let server = activeDnsServer, !providerIsMissing else {
            return String.localizedString("system_dns_server")
        }
        
        let provider = activeDnsProvider
        
        return createServerName(server: server, provider: provider)
    }
    
    func getServer(serverId: Int) -> DnsServerInfo? {
        for provider in allProviders {
            for server in provider.servers ?? [] {
                if Int(server.serverId) == serverId {
                    return server
                }
            }
        }
        
        return nil
    }
    
    func getServerName(serverId: Int) -> String? {
        for provider in allProviders {
            for server in provider.servers ?? [] {
                if Int(server.serverId) == serverId {
                    return createServerName(server: server, provider: provider)
                }
            }
        }
        
        return nil
    }
    
    func reset() {
        customProvidersInternal = nil
        predefinedProvidersInternal = nil
    }
    
    var adguardDohServer: DnsServerInfo? {
        return serverWithId("adguard-doh")
    }
    
    var adguardFamilyDohServer: DnsServerInfo? {
        return serverWithId("adguard-family-doh")
    }
    
    // MARK: - private methods
    
    private func createServerName(server: DnsServerInfo, provider: DnsProviderInfo?)->String {
        if isCustomServer(server) {
            return provider?.name ?? server.name
        }
        
        let protocolName = String.localizedString(DnsProtocol.stringIdByProtocol[server.dnsProtocol]!)
        return "\(provider?.name ?? server.name) (\(protocolName))"
    }
    
    private func serverWithId(_ id: String)->DnsServerInfo? {
        for provider in allProviders {
            for server in provider.servers ?? [] {
                if server.serverId == id {
                    return server
                }
            }
        }
        return nil
    }
    
    func featuresFromArray(_ arr: [String], featuresMap: [String: DnsProviderFeature]) -> [DnsProviderFeature] {
        var features = [DnsProviderFeature]()
        
        for featureName in arr {
            features.append(featuresMap[featureName]!)
        }
        
        return features
    }
    
    private func serversFromArray(_ arr:[[String: Any]]) -> [DnsServerInfo] {
        
        var servers = [DnsServerInfo]()
        
        for serverJson in arr {
            guard   let serverProtocolId = serverJson["type"] as? String,
                    let serverId = serverJson["id"] as? String,
                    let name = serverJson["name"] as? String,
                    let upstreams = serverJson["upstreams"] as? [String]
                else { continue }
            
            let server = DnsServerInfo(dnsProtocol: DnsProtocol.protocolByString[serverProtocolId]!, serverId: serverId, name: name, upstreams: upstreams)
            
            servers.append(server)
        }
        
        return servers
    }
    
    private func defaultProtocol(_ provider: DnsProviderInfo) -> DnsProtocol {
        
        let doh = provider.servers?.first { (dns) -> Bool in
            return dns.dnsProtocol == DnsProtocol.doh
        }
        
        if doh != nil {
            return .doh
        }
        
        let dot = provider.servers?.first { (dns) -> Bool in
            return dns.dnsProtocol == DnsProtocol.dot
        }

        if dot != nil {
            return .dot
        }
        
        let dnsCrypt = provider.servers?.first { (dns) -> Bool in
            return dns.dnsProtocol == DnsProtocol.dnsCrypt
        }
        
        if dnsCrypt != nil {
            return .dnsCrypt
        }
        
        let regular = provider.servers?.first { (dns) -> Bool in
            return dns.dnsProtocol == DnsProtocol.dns
        }
        
        if regular != nil {
            return .dns
        }
        
        return provider.servers?.first?.dnsProtocol ?? .dns
    }
    
    /*
     Check if provider exists for selected server
     If not, set active server to nil
     */
    private func missingProviderCheck() {
        if !providerIsMissing { return }
        DDLogInfo("Setting active dns server to nil; Current active server id: \(activeDnsServer?.serverId ?? "nil")")
        activeDnsServer = nil
        vpnManager?.updateSettings(completion: nil)
    }
    
    /* Initializes providers from providers.json */
    private func initializeDnsProviders() {
        guard  let jsonData = getData(forFileName: "providers") else {
            DDLogError("Got nil json data")
            return
        }
        
        let decoder = JSONDecoder()
        do {
            let decodedObject = try decoder.decode(DnsProviders.self, from: jsonData)
            let dnsProviders = decodedObject.providers.filter { (provider) -> Bool in
                provider.providerId != systemDefaultProviderId
            }
            let features = decodedObject.features
            localize(dnsProviders: dnsProviders, features: features)
        }
        catch {
            DDLogError("Failed to parse providers.json; error: \(error.localizedDescription)")
        }
    }
    
    /* Gets localizations for providers from providers_i18n.json */
    private func localize(dnsProviders: [DnsProvider], features: [DnsFeature]) {
        guard  let jsonData = getData(forFileName: "providers_i18n") else {
            DDLogError("Got nil json data")
            return
        }
        
        var json: [String: Any] = [:]
        do {
            json = try JSONSerialization.jsonObject(with: jsonData) as! [String: Any]
        } catch {
            DDLogError(error.localizedDescription)
            return
        }
        
        // Fill providers from json
        let dnsProviders = getLocalizedProvidersForCurrentLocale(dnsProviders: dnsProviders, features: features, localizationsJson: json)
        
        self.predefinedProvidersInternal = dnsProviders.map{ (provider) -> DnsProviderInfo in
            let providerInfo = DnsProviderInfo(name: provider.localizedName ?? "")
            
            providerInfo.logo = provider.provider.logo
            providerInfo.logoDark = "\(provider.provider.logo)_dark"
            providerInfo.protocols = provider.provider.servers.map { DnsProtocol(type: $0.type) }
            providerInfo.summary = provider.localizedDescription
            
            providerInfo.features = provider.features?.map {
                DnsProviderFeature(name: $0.featureId, title: $0.localizedName ?? "", summary: $0.localizedDescription ?? "", iconId: $0.logo ?? "")
            }
            
            providerInfo.servers = provider.provider.servers.map {
                return DnsServerInfo(dnsProtocol: DnsProtocol(type: $0.type), serverId: String($0.id), name: "", upstreams: $0.upstreams)
            }
            
            let protcol = providerInfo.getActiveProtocol(resources)
            if protcol == nil {
                let prot = defaultProtocol(providerInfo)
                providerInfo.setActiveProtocol(resources, protcol: prot)
            }
            
            return providerInfo
        }
    }
    
    /* Returns json Data from file name */
    private func getData(forFileName fileName: String) -> Data? {
        guard let pathString = Bundle.main.path(forResource: fileName, ofType: "json") else {
            DDLogError("Failed to get \(fileName).json path")
            return nil
        }
        
        let pathUrl = URL(fileURLWithPath: pathString)
        
        guard let jsonData = try? Data(contentsOf: pathUrl) else {
            DDLogError("Failed to get json data from path: \(pathString)")
            return nil
        }
        
        return jsonData
    }
    
    /* Fills dns providers with localizations for current locale */
    private func getLocalizedProvidersForCurrentLocale(dnsProviders: [DnsProvider], features: [DnsFeature], localizationsJson: [String: Any]) -> [LocalizedDnsProviderInfo] {
        guard let providersJson = localizationsJson["providers"] as? [String: Any] else {
            DDLogError("There is no field 'providers' in providers_i18n json")
            return []
        }
        
        guard let featuresLocalizationsJson = localizationsJson["features"] as? [String: Any] else {
            DDLogError("There is no field 'features' in providers_i18n json")
            return []
        }
        
        let localizedDnsProviders = dnsProviders.compactMap { dnsProvider -> LocalizedDnsProviderInfo? in
            return localize(dnsProvider: dnsProvider, features: features, fromJson: providersJson, featuresLocalizationsJson: featuresLocalizationsJson)
        }
        
        return localizedDnsProviders
    }
    
    /* Localizes dns provider from json with localizations */
    private func localize(dnsProvider: DnsProvider, features:[DnsFeature], fromJson json: [String: Any], featuresLocalizationsJson: [String: Any]) -> LocalizedDnsProviderInfo? {
        let localeCode = languageMap[currentLocaleCode] ?? currentLocaleCode
        
        let providerId = String(dnsProvider.providerId)
        
        guard let providerLocalizationsJson = json[providerId] as? [String: Any] else {
            DDLogError("There is no json for provider with id = \(providerId)")
            return nil
        }
        
        let translation = getTranslation(fromJson: providerLocalizationsJson, forLocale: localeCode)
        
        let name = translation?.name
        let description = translation?.description
        
        var featuresSet = Set<String>()
        for server in dnsProvider.servers {
            for feature in server.features {
                featuresSet.insert(feature)
            }
        }
        
        let featuresArray = Array<String>(featuresSet)
        
        var featureLogos = [String:String]()
        for featureInfo in features {
            featureLogos[featureInfo.type] = featureInfo.logo
        }
        
        let localizedFeatures = featuresArray.compactMap { (feature) -> LocalizedDnsFeature? in
            guard let featureJson = featuresLocalizationsJson[feature] as? [String: Any] else {
                DDLogError("There is no localization for reature with id = \(feature)")
                return nil
            }
            let translation = getTranslation(fromJson: featureJson, forLocale: localeCode)
            let localizedFeature = LocalizedDnsFeature(featureId: feature, name: translation?.name, description: translation?.description, logo: featureLogos[feature])
            return localizedFeature
        }
        
        return LocalizedDnsProviderInfo(provider: dnsProvider, name: name, description: description, features: localizedFeatures)
    }
    
    private func getTranslation(fromJson json: [String: Any], forLocale locale: String) -> (name: String?, description: String?)? {
        // Get english translations
        guard let enLocaleJson = json["en"] as? [String: Any] else {
            DDLogError("English description for provider is missing")
            return nil
        }
        let enName = enLocaleJson["name"] as? String
        let enDescription = enLocaleJson["description"] as? String
        
        guard let currentLocaleJson = json[locale] as? [String: Any] else {
            // If json for current locale is missing return english translations
            return (enName, enDescription)
        }
        
        let nameForLocale = currentLocaleJson["name"] as? String
        let descriptionForLocale = currentLocaleJson["description"] as? String
        
        return (nameForLocale ?? enName, descriptionForLocale ?? enDescription)
    }
}
